unit BlobCds;

 {Version 1.03}
 {Amazing Component No. 3 bY Jirayu Wiriyaphibool (Software Mission) on Sept 15. 2000}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, DB,
  DBCtrls, DBClient, Variants, CompressEx, Dialogs;

type

  TBlobCds = class;

  TOnAfterLoad   = procedure(DataSet: TDataSet) of object;
  TOnDefineField = procedure(DataSet: TDataSet) of object;
  TOnDefineFieldLabel = procedure(DataSet: TDataSet) of object;

  TBlobCds = class(TClientDataSet)
  private
    CloneDat:      TClientDataSet;
    FReadyToUpdate: boolean;
    FOnAfterLoad:  TOnAfterLoad;
    FOnDefineField: TOnDefineField;
    FOnDefineFieldLabel: TOnDefineFieldLabel;
    FDataLink:     TFieldDataLink;
    FModified:     boolean;
    FAllowUndo:    boolean;
    FCompressed:   boolean;
    FFieldVisibles: TStrings;
    FFieldLabels:  TStrings;
    FFieldWidths:  TStrings;
    FFieldFormats: TStrings;
    FFieldSequence: string;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    function GetDataField: string;
    function GetDataSource: TDataSource;
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    procedure ActiveChange(Sender: TObject);
    procedure UpdateData(Sender: TObject);
    function SetEditing: boolean;
    procedure SetAllowUndo(const Value: boolean);
    function GetField(const FieldName: string): variant;
    procedure SetField(const FieldName: string; const Value: variant);
    procedure SetFieldWidth(FieldName: string; Len: integer);
    procedure SetFieldFormats(const Value: TStrings);
    procedure SetFieldLabels(const Value: TStrings);
    procedure SetFieldVisibles(const Value: TStrings);
    procedure SetFieldWidths(const Value: TStrings);
    procedure DetermineFieldAttributes;
    procedure SetFieldSequence(const Value: string);
    procedure SaveToStreamOrdered(Stream: TStream);
  public
    HasBeenModified: boolean;
    MemData: TStringStream;
    MemOut:  TStringStream;
    procedure DoAfterOpen; override;
    procedure DoBeforeEdit; override;
    procedure DoBeforeDelete; override;
    procedure DoBeforeInsert; override;
    procedure ReLoad;
    procedure LoadData(FieldName: string; const SortFields: string = '');
    procedure SaveData(FieldName: string); overload;
    procedure SaveData; overload;
    procedure KeepContent;
    procedure RestoreContent;
    procedure AddField(FieldName: string; DataType: TFieldType; Size: integer;
      const Position: integer = 0);
    procedure AddCalcField(FieldName: string; DataType: TFieldType; Len: integer;
      const Position: integer = 0);
    procedure DeleteField(FieldName: string);
    constructor Create(AOwner: TComponent); override;
    procedure Loaded; override;
    destructor Destroy; override;
    procedure SetLabels(LabelList: string);
    procedure SetPos(PositionList: string);
    procedure SetWidth(WidthList: string);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function SumField(FieldName: string): extended;
    procedure UpGradeFloat(OldName, NewName: string);
    procedure Wipe;
    function IsEmpty: boolean;
    procedure LoadFromField(DataSet: TDataset; FieldName: string);
    procedure SaveToField(DataSet:TDataSet;FieldName:String);
    procedure FieldSeq(ListofFields: string);
    property Field[const FieldName: string]: variant Read GetField Write SetField;
      default;
  published
    property FieldLabels: TStrings Read FFieldLabels Write SetFieldLabels;
    property FieldWidths: TStrings Read FFieldWidths Write SetFieldWidths;
    property FieldVisibles: TStrings Read FFieldVisibles Write SetFieldVisibles;
    property FieldFormats: TStrings Read FFieldFormats Write SetFieldFormats;
    property FieldSequence: string Read FFieldSequence Write SetFieldSequence;
    property AllowUndo: boolean Read FAllowUndo Write SetAllowUndo;
    property DataField: string Read GetDataField Write SetDataField;
    property Compressed: boolean Read FCompressed Write FCompressed;
    property DataSource: TDataSource Read GetDataSource Write SetDataSource;
    property OnDefineField: TOnDefineField Read FOnDefineField Write FOnDefineField;
    property OnAfterLoad: TOnAfterLoad Read FOnAfterLoad Write FOnAfterLoad;
    property OnDefineFieldLabel: TOnDefineFieldLabel
      Read FOnDefineFieldLabel Write FOnDefineFieldLabel;
  end;

implementation

var
  FakeField: TField;

{ TBlobCds }

function GetStr(var Inpstr: string; Delim: char): string;
var
  i: integer;
begin
  i := Pos(Delim, InpStr);
  if i = 0 then
  begin
    Result := Inpstr;
    Inpstr := '';
  end
  else
  begin
    Result := Copy(InpStr, 1, i - 1);
    Delete(InpStr, 1, I);
  end;
end;

function LastPos(InpStr, s: string): byte;
var
  i, lsub, ls: integer;
begin
  LastPos := 0;
  ls      := Length(s);
  lsub    := Length(InpStr);

  if (ls > 0) and (lsub > 0) then
    for i := ls - lsub + 1 downto 1 do
    begin
      if InpStr = Copy(s, i, lsub) then
      begin
        LastPos := i;
        Exit;
      end;
    end;
end;

function Between(SubStr1, SubStr2, s: string): string;
var
  i, j: integer;
begin
  Between := '';
  i := Pos(SubStr1, s);
  if i <> 0 then
  begin
    i := i + Length(SubStr1);
    j := LastPos(SubStr2, s);
    if (j <> 0) and (j > i) then
      Between := Copy(s, i, j - i);
  end;
end;

function Before(SubStr, s: string): string;
var
  i: integer;
begin
  i := Pos(SubStr, s);
  if i > 0 then
    Before := Copy(s, 1, i - 1)
  else
    Before := '';
end;

function After(InpStr, s: string): string;
var
  i: integer;
begin
  i := Pos(InpStr, s);
  if i <> 0 then
    Result := Copy(s, i + Length(InpStr), Length(S));
end;

procedure TBlobCds.SaveToStreamOrdered(Stream: TStream);
var
  CopyCds: TBlobCds;
  CurrentRecNo, FldCount, Cnt: integer;
begin
  CheckBrowseMode;
  CopyCds := TBlobCds.Create(Self);
  try
    CopyCds.FieldDefs.Assign(FieldDefs);
    CopyCds.CreateDataSet;
    FldCount     := FieldDefs.Count;
    CurrentRecNo := RecNo;
    try
      DisableControls;
      First;
      while not EOF do
      begin
        CopyCds.Insert;
        for Cnt := FldCount - 1 downto 0 do
          if not (FieldList[Cnt].IsNull) then
            CopyCds.FieldList[Cnt].Value := FieldList[Cnt].Value
          else
            CopyCds.FieldList[Cnt].Clear;
        Next;
      end;
      MergeChangeLog;
      CopyCds.SaveToStream(Stream);
    finally
      RecNo := CurrentRecNo;
      EnableControls;
    end;
  finally
    CopyCds.Free;
  end;
end;

constructor TBlobCds.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFieldLabels   := TStringList.Create;
  FFieldWidths   := TStringList.Create;
  FFieldVisibles := TStringList.Create;
  FFieldFormats  := TStringList.Create;
  FAllowUndo     := False;
  FDataLink      := TFieldDataLink.Create;
  FDataLink.Control := Self;
  FDataLink.OnDataChange := DataChange;
  FDataLink.OnEditingChange := EditingChange;
  FDataLink.OnUpdateData := UpdateData;
  FDataLink.OnActiveChange := ActiveChange;
end;

procedure TBlobCds.DataChange(Sender: TObject);
begin
  if (csDesigning in ComponentState) then
    Exit;
  if (csDestroying in ComponentState) or (csLoading in ComponentState) then
    Exit;
  if (FDataLink.DataSet = nil) then
    Exit;
  if (FDataLink.Field = nil) then
    Exit; // don't ever remove this line, damn it.
  if (FModified) and (FDataLink.DataSet.State <> dsInsert) then
    Exit;
  Reload; // don't load when form not ready (loaded hasnt been called or AV)
end;


function TBlobCds.SetEditing: boolean;
begin
  if (FDataLink.CanModify) then
  begin
    if not (FDataLink.Editing) then
      FDatalink.Edit;
    if FDataLink.Editing then
    begin
      FDataLink.Modified;
      Result := True;
    end
    else
      Result := False;
  end;
end;


procedure TBlobCds.EditingChange(Sender: TObject);
begin
  if (FDatalink.editing) then
    FModified := True
  else
    FModified := False;
end;


destructor TBlobCds.Destroy;
begin
  FFieldLabels.Free;
  FFieldWidths.Free;
  FFieldVisibles.Free;
  FFieldFormats.Free;
  FDataLink.OnDataChange := nil;
  FDataLink.OnEditingChange := nil;
  FDataLink.OnUpdateData := nil;
  FDataLink.Control := nil;
  FDataLink.Free;
  FDataLink := nil;
  inherited Destroy;
end;


function TBlobCds.GetDataField: string;
begin
  Result := FDataLink.FieldName;
end;

function TBlobCds.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TBlobCds.SetDataField(const Value: string);
begin
  FDataLink.FieldName := Value;
end;

procedure TBlobCds.SetDataSource(Value: TDataSource);
begin
  if FDataLink.DataSource <> Value then
  begin
    FDataLink.DataSource := Value;
  end;
  if Value <> nil then
    Value.FreeNotification(Self);
end;

procedure TBlobCds.UpdateData(Sender: TObject);
begin
  try
    MemData := TStringStream.Create('');
    MemOut  := TStringStream.Create('');
    if not (FAllowUndo) then
      MergeChangeLog;
    if FCompressed then
    begin
      SaveToStreamOrdered(MemData);
      LZCompressBlobStream(MemData, MemOut);
    end
    else
      SaveToStreamOrdered(MemOut);
    MemOut.Seek(0, SoFromBeginning);
    TBlobField(FDataLink.Field).LoadFromStream(MemOut);
  finally
    MemData.Free;
    MemOut.Free;
  end;
end;

procedure TBlobCds.DoBeforeEdit;
begin
  inherited DoBeforeEdit;
  if not (SetEditing) then
    Exit;
  HasBeenModified := True;
end;

procedure TBlobCds.DoBeforeDelete;
begin
  inherited DoBeforeDelete;
  SetEditing;
  HasBeenModified := True;
end;

procedure TBlobCds.ReLoad;
begin
  if FDataLink = nil then
    Exit;//crucial for manual usage
  if FDataLink.Field = nil then
    Exit; //crucial for manual usage (ASSIGN DATASOURCE BUT NOT BLOB FIELD
  HasBeenModified := False;
  if not (TBlobField(FDataLink.Field).IsNull) then
  begin
    try
      MemData := TStringStream.Create('');
      MemOut  := TStringStream.Create('');
      TBlobField(FDataLink.Field).SaveToStream(MemData);
      Memdata.Seek(0, SoFromBeginning);
      IndexFieldNames := '';
      if FCompressed then
      begin
        LZDecompressBlobStream(MemData, MemOut);
        MemOut.Seek(0, SofromBeginning);
        LoadFromStream(MemOut);
      end
      else
        LoadFromStream(MemData);
      if Assigned(FOnAfterLoad) then
        FOnAfterLoad(Self);
      if Assigned(FOnDefineFieldLabel) then
        FOnDefineFieldLabel(Self);
    finally
      MemOut.Free;
      MemData.Free;
      First;
    end;
  end
  else
  begin
    if Assigned(FOnDefineField) then
    begin
      Close;
      FieldDefs.Clear;
      FOnDefineField(Self);
      CreateDataSet;
      if Assigned(FOnAfterLoad) then
        FOnAfterLoad(Self);
      if Assigned(FOnDefineFieldLabel) then
        FOnDefineFieldLabel(Self);
      LogChanges := False;
    end
    else
    begin
      Close;
      CreateDataSet;
      if Assigned(FOnAfterLoad) then
        FOnAfterLoad(Self);
      if Assigned(FOnDefineFieldLabel) then
        FOnDefineFieldLabel(Self);
    end;
  end; // else
end;

procedure TBlobCds.DoBeforeInsert;
begin
  inherited DoBeforeInsert;
  if not (SetEditing) then
    Exit;
  HasBeenModified := True;
end;

procedure TBlobCds.SetAllowUndo(const Value: boolean);
begin
  FAllowUndo := Value;
end;

procedure TBlobCds.SaveData;
begin
  FDataLink.DataSet.Edit;
  UpdateData(Self);
  FDataLink.DataSet.Post;
end;

procedure TBlobCds.Loaded;
begin
  inherited Loaded;
end;

procedure TBlobCds.RestoreContent;
var
  I: integer;
  F: TField;
begin
  if not (FReadyToUpdate) then
    Exit;
  DisableControls;
  Close;
  ObjectView := True;
  // Require to support nested fields this will flat out all fields
  // so it is easy to transfer content. Nothing beats clientdataset for this kind of stuff
  CreateDataSet;
  CloneDat.First;
  while not (CloneDat.EOF) do
  begin
    Append;
    for I := 0 to CloneDat.Fields.Count - 1 do
    begin
      F := FindField(CloneDat.Fields[I].FieldName);
      if (F <> nil) and (F.DataType <> ftAutoInc) and (F.FieldKind <> fkCalculated) and
        (F.DataType <> ftDataSet) then
      begin
        F.Value := CloneDat.Fields[I].Value; // * Now support multi level
      end;
    end;
    Post;
    CloneDat.Next;
  end;
  CloneDat.Data := Data; // Keep it one more time;
  Close;
  ObjectView := True; // This one to support nested field
  Data := CloneDat.Data;
  First;
  FReadyToUpdate := False;
  EnableControls;
end;

procedure TBlobCds.ActiveChange(Sender: TObject);
begin
  if FDataLink.Active then
    if FDataLink.DataSet.EOF <> FDatalink.DataSet.Bof then
      Reload;
end;

procedure TBlobCds.DoAfterOpen;
var
  I, Where: integer;
  S: string;
begin
  inherited;
  for I := 0 to Fields.Count - 1 do
  begin
    repeat
      Where := Pos('_', Fields[I].DisplayLabel);
      if Where <> 0 then
      begin
        S := Fields[I].DisplayLabel;
        System.Delete(S, Where, 1);
        System.Insert(' ', S, Where);
        Fields[I].DisplayLabel := S;
      end;
    until Where = 0;
  end;
  if Assigned(FOnDefineFieldLabel) then
    FOnDefineFieldLabel(self);
  if Trim(FFieldSequence) <> '' then
    FieldSeq(FFieldSequence);
  DetermineFieldAttributes;
end;

procedure TBlobCds.KeepContent;
begin
  if CloneDat = nil then
    CloneDat := TClientDataSet.Create(Self);

  CloneDat.Close;
  CloneDat.ObjectView := True;
  CloneDat.Data  := Data;
  FReadyToUpdate := True;
end;

function TBlobCds.GetField(const FieldName: string): variant;
var
  Fld: TFieldType;
begin
  if not (FieldByName(FieldName).IsNull) then
    Result := FieldByName(FieldName).Value
  else
  begin
    Fld := FieldByName(FieldName).DataType;
    case Fld of
      ftString, ftMemo: Result := FieldByName(FieldName).AsString;
      ftInteger: Result := FieldByName(FieldName).AsInteger;
      ftFloat: Result   := FieldByName(FieldName).AsFloat;
      ftCurrency, ftBCD: Result := FieldByName(FieldName).AsCurrency;
      ftBoolean: Result := FieldByName(FieldName).AsBoolean;
      ftDateTime, ftTime: Result := FieldByName(FieldName).AsDateTime;
      ftVariant: Result := FieldByName(FieldName).AsVariant;
    end;
  end;
end;

procedure TBlobCds.SetField(const FieldName: string; const Value: variant);
begin
  if Value <> Null then
    FieldByName(FieldName).Value := Value;
end;

procedure TBlobcds.SetLabels(LabelList: string);
var
  ThatField, FieldName, LabelName: string;
  Where: integer;
begin
  repeat
    ThatField := GetStr(LabelList, ',');
    Where     := Pos('=', ThatField);
    if Where <> 0 then
    begin
      FieldName := Copy(ThatField, 1, Where - 1);
      LabelName := Copy(ThatField, Where + 1, Length(ThatField));
      if FindField(FieldName) <> nil then
        FieldByName(FieldName).DisplayLabel := LabelName;
    end;
  until LabelList = '';
end;

procedure TBlobcds.SetPos(PositionList: string);
var
  ThatField, FieldName, IndexPos: string;
  Where: integer;
begin
  repeat
    ThatField := GetStr(PositionList, ',');
    Where     := Pos('=', ThatField);
    if Where <> 0 then
    begin
      FieldName := Copy(ThatField, 1, Where - 1);
      IndexPos  := Copy(ThatField, Where + 1, Length(ThatField));
      if FindField(FieldName) <> nil then
        FieldByName(FieldName).Index := StrToInt(Trim(IndexPos));
    end;
  until PositionList = '';
end;

procedure TBlobcds.SetFieldWidth(FieldName: string; Len: integer);
begin
  if FindField(FieldName) <> nil then
    FieldByName(FieldName).DisplayWidth := Len;
end;

procedure TBlobcds.SetWidth(WidthList: string);
var
  ThatField, FieldName, FieldWidth: string;
  Where: integer;
begin
  repeat
    ThatField := GetStr(WidthList, ',');
    Where     := Pos('=', ThatField);
    if Where <> 0 then
    begin
      FieldName  := Copy(ThatField, 1, Where - 1);
      FieldWidth := Copy(ThatField, Where + 1, Length(ThatField));
      SetFieldWidth(FieldName, StrToInt(Trim(FieldWidth)));
    end;
  until WidthList = '';
end;

procedure TBlobCds.AddField(FieldName: string; DataType: TFieldType; Size: integer;
  const Position: integer = 0);
var
  F: TFieldDef;
begin
  KeepContent;
  if FieldDefs.IndexOf(FieldName) <> -1 then
    Exit;
  FieldDefs.Update;
  FieldDefs.Add(FieldName, DataType, Size, False);
  F := FieldDefs.Find(FieldName);
  F.FieldNo := 0;
  if Position <> 0 then
    F.Index := Position;
  RestoreContent;
end;

procedure TBlobCds.DeleteField(FieldName: string);
var
  I, N:  integer;
  Token: string;
  F:     TField;
begin
  F := FindField(FieldName);
  if F <> nil then
  begin
    KeepContent;
    if FieldName <> '' then
    begin
      N := F.FieldNo - 1;
      Close;
      FieldDefs.Update;
      FieldDefs.Delete(N);
    end;
    RestoreContent;
  end;
end;

function TBlobCds.SumField(FieldName: string): extended;
var
  GrandAmt: extended;
  DS: TClientDataSet;
begin
  if FindField(FieldName) = nil then
  begin
    Result := 0.00;
    Exit;
  end;
  DS      := TClientDataSet.Create(nil);
  DS.Data := Data;
  GrandAmt := 0;
  DS.First;
  while not (DS.EOF) do
  begin
    if not (DS.FieldByName(FieldName).IsNull) then
      GrandAmt := GrandAmt + DS.FieldByName(FieldName).AsFloat;
    DS.Next;
  end;
  DS.Free;
  Result := GrandAmt;
end;


procedure TBlobCds.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = OpRemove) and (FDataLink <> nil) then
    if AComponent = Datasource then
      DataSource := nil;
end;

procedure TBlobCds.Wipe;
begin
  DisableControls;
  try
    First;
    while not (EOF) do
      Delete;
  finally
    EnableControls;
  end;
end;

procedure TBlobCds.LoadFromField(DataSet: TDataset; FieldName: string);
 //var B:TBlobField;
 //Mem:TMemoryStream;
var
  F: TBlobField;
begin
  F := TBlobField(DataSet.FieldByName(FieldName));
  if not (F.IsNull) then
  begin
    try
      MemData := TStringStream.Create('');
      MemOut  := TStringStream.Create('');
      F.SaveToStream(MemData);
      Memdata.Seek(0, SoFromBeginning);
      if FCompressed then
      begin
        LZDecompressBlobStream(MemData, MemOut);
        MemOut.Seek(0, SofromBeginning);
        LoadFromStream(MemOut);
      end
      else
        LoadFromStream(MemData);
      if Assigned(FOnAfterLoad) then
        FOnAfterLoad(Self);
      if Assigned(FOnDefineFieldLabel) then
        FOnDefineFieldLabel(Self);
    finally
      MemOut.Free;
      MemData.Free;
      First;
    end;
  end
 else
  begin
     if FieldDefs.Count > 0 then
        begin
          close;
          CreateDataSet;
        end;  
  end;
end;

procedure TBlobCds.LoadData(FieldName: string; const SortFields: string = '');
var
  F: TBlobField;
begin
  if FDataLink.DataSet = nil then
    Exit;
  F := TBlobField(FDataLink.DataSet.FieldByName(FieldName));
  if F <> nil then
  begin
    HasBeenModified := False;
    if not (F.IsNull) then
    begin
      try
        MemData := TStringStream.Create('');
        MemOut  := TStringStream.Create('');
        F.SaveToStream(MemData);
        Memdata.Seek(0, SoFromBeginning);
        if SortFields <> '' then
          IndexFieldNames := SortFields
        else
          IndexFieldNames := '';
        if FCompressed then
        begin
          LZDecompressBlobStream(MemData, MemOut);
          MemOut.Seek(0, SofromBeginning);
          LoadFromStream(MemOut);
        end
        else
          LoadFromStream(MemData);
        if Assigned(FOnAfterLoad) then
          FOnAfterLoad(Self);
        if Assigned(FOnDefineFieldLabel) then
          FOnDefineFieldLabel(Self);
      finally
        MemOut.Free;
        MemData.Free;
        First;
      end;
    end
    else
    begin
      if Assigned(FOnDefineField) then
      begin
        Close;
        FieldDefs.Clear;
        FOnDefineField(Self);
        CreateDataSet;
        if Assigned(FOnAfterLoad) then
          FOnAfterLoad(Self);
        if Assigned(FOnDefineFieldLabel) then
          FOnDefineFieldLabel(Self);
        //LogChanges := False;
      end
      else
      begin
        Close;
        CreateDataSet;
        if Assigned(FOnDefineFieldLabel) then
          FOnDefineFieldLabel(Self);
        if Assigned(FOnAfterLoad) then
          FOnAfterLoad(Self);
      end;
    end; // else
  end;
end;

procedure TBlobCds.SaveData(FieldName: string);
var
  F: TBlobField;
begin
  if FDataLink.DataSet = nil then
    Exit;
  F := TBlobField(FDataLink.DataSet.FieldByName(FieldName));
  if F <> nil then
  begin
    FDataLink.DataSet.Edit;
    try
      MemData := TStringStream.Create('');
      MemOut  := TStringStream.Create('');
      if not (FAllowUndo) then
        MergeChangeLog;
      if FCompressed then
      begin
        SaveToStream(MemData);
        LZCompressBlobStream(MemData, MemOut);
      end
      else
        SaveToStream(MemOut);
      MemOut.Seek(0, SoFromBeginning);
      F.LoadFromStream(MemOut);
    finally
      MemData.Free;
      MemOut.Free;
    end;
    FDataLink.DataSet.Post;
  end;
end;

procedure TBlobCds.SetFieldFormats(const Value: TStrings);
begin
  FFieldFormats.Assign(Value);
end;

procedure TBlobCds.SetFieldLabels(const Value: TStrings);
begin
  FFieldLabels.Assign(Value);
end;

procedure TBlobCds.SetFieldVisibles(const Value: TStrings);
begin
  FFieldVisibles.Assign(Value);
end;

procedure TBlobCds.SetFieldWidths(const Value: TStrings);
begin
  FFieldWidths.Assign(Value);
end;

procedure TBlobCds.FieldSeq(ListofFields: string);
var
  Idx:     integer;
  FldName: string;
begin
  Idx := 0;
  repeat
    FldName := GetStr(ListOfFields, ',');
    if Trim(FldName) <> '' then
    begin
      if FindField(FldName) <> nil then
      begin
        FieldByName(FldName).Index := Idx;
        Inc(Idx);
      end;
    end;
  until FldName = '';
end;

procedure TBlobCds.DetermineFieldAttributes;
var
  I: integer;
  F: TField;
  Key, S, Value: string;
begin
  for I := 0 to FFieldFormats.Count - 1 do
  begin
    S     := FFieldFormats[I];
    Key   := Before('=', S);
    Value := After('=', S);
    F     := FindField(Key);
    if F <> nil then
    begin
      if (F is TNumericField) then
        TNumericField(F).DisplayFormat := Value;
      if (F is TDateField) then
        TDateField(F).DisplayFormat := Value;
    end;
  end;
  for I := 0 to FFieldWidths.Count - 1 do
  begin
    S     := FFieldWidths[I];
    Key   := Before('=', S);
    Value := After('=', S);
    F     := FindField(Key);
    if F <> nil then
      F.DisplayWidth := StrToInt(Value);
  end;
  for I := 0 to FFieldLabels.Count - 1 do
  begin
    S     := FFieldLabels[I];
    Key   := Before('=', S);
    Value := After('=', S);
    F     := FindField(Key);
    if F <> nil then
      F.DisplayLabel := Value;
  end;
  for I := 0 to FFieldVisibles.Count - 1 do
  begin
    S     := FFieldVisibles[I];
    Key   := Before('=', S);
    Value := After('=', S);
    F     := FindField(Key);
    if F <> nil then
    begin
      if Value = '0' then
        F.Visible := False;
      if Value = '1' then
        F.Visible := True;
    end;
  end;
end;


procedure TBlobCds.SetFieldSequence(const Value: string);
begin
  FFieldSequence := Value;
end;

procedure TBlobCds.AddCalcField(FieldName: string; DataType: TFieldType;
  Len: integer; const Position: integer);
var
  I: integer;
begin
  if FindField(FieldName) <> nil then
    Exit;
  FieldDefs.Update;
  Close;
  for I := 0 to FieldDefs.Count - 1 do
  begin
    try
      FieldDefs[i].CreateField(self);
    except
    end;
  end;
  case DataType of
    ftMemo: FakeField := TMemoField.Create(self);
    ftString:
    begin
      FakeField      := TStringField.Create(self);
      FakeField.Size := Len;
    end;
    ftInteger: FakeField  := TIntegerField.Create(self);
    ftFloat: FakeField    := TFloatField.Create(self);
    ftDateTime: FakeField := TDateTimeField.Create(self);
    ftDate: FakeField     := TDateField.Create(self);
  end;
  FakeField.FieldName := FieldName;
  FakeField.FieldKind := fkCalculated;
  FakeField.DataSet   := self;
  FakeField.Index     := Position;
end;

function TBlobCds.IsEmpty: boolean;
begin
  if Active then
  begin
    if TBlobField(FDataLink.Dataset.FieldByName(FDataLink.FieldName)).IsNull then
    begin
      Result := True;
      Exit;
    end
    else
      Result := False;
    if (BOF = EOF) then
      Result := True
    else
      Result := False;
  end
  else
    raise Exception.Create('Blob CDS not active');
end;

procedure TBlobCds.UpGradeFloat(OldName, NewName: string);
var
  F:  TField;
  TN: string;
begin
  F := FindField(OldName);
  if F <> nil then
  begin
    TN := OldName + '_TM';
    if F.DataType = ftInteger then
    begin
      AddField(TN, ftFloat, 0);
      First;
      while not (EOF) do
      begin
        Edit;
        Field[TN] := Field[OldName];
        Next;
      end;
      DeleteField(OldName);
      AddField(NewName, ftFloat, 0);
      First;
      while not (EOF) do
      begin
        Edit;
        Field[NewName] := Field[TN];
        Next;
      end;
      DeleteField(TN);
      ReLoad;
    end;
  end;
end;

procedure TBlobCds.SaveToField(DataSet: TDataSet; FieldName: String);
var F:TBlobField;
begin
  F := TBlobField(DataSet.FieldByName(FieldName));
  try
    if FCompressed then
    begin
      MemData := TStringStream.Create('');
      MemOut  := TStringStream.Create('');
      SaveToStream(MemData);
      LZCompressBlobStream(MemData, MemOut);
      MemOut.Seek(0, SofromBeginning);
      Edit;
      F.LoadFromStream(MemOut);
      Post;
    end
    else
    begin
      MemData := TStringStream.Create('');
      MemOut  := TStringStream.Create('');
      SaveToStream(MemData);
      MemData.Seek(0,SofromBeginning);
      Edit;
      F.LoadFromStream(MemData);
      Post;
    end;
  finally
    MemOut.Free;
    MemData.Free;
  end;
end;

end.
